#include <memory>
#include <vector>

/**
 * This shall declare and define a signal flow graph based on the current 
 * state of the art in both signal processing software engineering and 
 * C++ programming. The graph shall be suitable both for Csound 6 and for 
 * Som, if that ever becomes more than a bag of ideas.
 *
 * My best guess is that the signal flow graph will be a heterogeneous 
 * multi-rate synchronous signal flow graph with an inital static 
 * schedule. At runtime, new instrument instantiations will cause the 
 * duplication of existing columns in the scheduling matrix and of 
 * existing actors in the graph. Just enough back-pointers will be 
 * added to the canonical graph to enable this.
 * 
 * Token -- An elementary signal type, or sample. In Csound, this would be an
 * audio sample, an audio transform coefficient, or a control signal sample.
 * 
 * FIFO -- A first-in, first-out queue of tokens.
 *
 * Data flow graph -- a directed graph consisting of actors connected 
 * by edges
 */

// Don't forget control/data flow graph -- which is maybe what we really have.
// I think the best Google search for the Csound engine is "schedule control data flow graph".

struct Node;
struct Outlet;
struct Inlet;
struct Polyphonic;

// Topology matrix for synchronous data flow graph G: Tb = 0 is the balance equation.
// T = rows of edges, columns of vertices, elements prd(a) if A = src(a), -cns(a) if A = snk(a).
// Then if each of b > 0, there is a static schedule for G; b is the solution of the balance equation.
// Obviously given a list of multirate opcodes, Eigen can solve for b and use T to run the graph.
// Furthermore, I'm pretty sure that simply creating a new row in T simply implies creating a copy
// of the corresponding element in b, so dynamic voice allocation will be OK. Then it's a question
// of putting back pointers in the elements of T.
// Look for "dataflow schedule graph" which seems to combine G and T. Also Maryland DSPCAD.

// To make all this work, we must mix untyped containers and base classes with typed ones. The upper
// layer of the graph, which is what actually runs, is untyped or generic. The lower layer of the 
// graph, which is what actually computes the actions, is typed so that the code knows what type 
// the tokens are.

class FIFO
{
	virtual void Push();
	virtual void Pop();
	virtual void *Front();
	virtual void *Back();
};

template<typename TOKEN, typename Container=std::deque>
class Fifo : public std::queue<TOKEN, std::deque>, public FIFO
{
};


struct Node
{
	std::vector<std::shared_ptr<Unit>> sources;
	std::vector<std::shared_ptr<Unit>> sinks;
	Unit()
	{
	}
	virtual ~Unit()
	{
	}
	virtual void open()
	{
	}
	virtual void start()
	{
	}
	virtual void tick(int32 step, int32 frames)
	{
	}
	virtual void process(int32 step, int32 frames)
	{
	}
	virtual void stop()
	{
	}
	virtual void close()
	{
	}
};